@@ -1,5 +1,5 @@
 /*
- * Copyright (c) [2016,2018] SUSE LLC
+ * Copyright (c) [2016-2019] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -115,6 +115,15 @@ ProxySnapshotsDbus::getDefault() const
 }
 
 
+ProxySnapshots::iterator
+ProxySnapshotsDbus::getDefault()
+{
+    pair<bool, unsigned int> tmp = command_get_default_snapshot(conn(), configName());
+
+    return tmp.first ? find(tmp.second) : end();
+}
+
+
 ProxySnapshots::const_iterator
 ProxySnapshotsDbus::getActive() const
 {
